
from jinja2 import Markup
import inspect
import copy
from collections import OrderedDict, deque

from djingles.html import Link
from django.utils.formats import localize
from djingles.utils import url_query_update, flatten

__all__ = ['Formatter', 'FormattedTable', 'FormattedObject']


class Formatter(object):

    __position = 1

    def __init__(self, label=None, attr=None, hidden=False, sortable=True, reverse=False, empty="",
                 icon=None, url=None, editable=False, order_by=None):
        Formatter.__position += 1
        self.__position = Formatter.__position
        self.label = label
        self.icon = icon
        self.url = url
        self.hidden = hidden
        self.editable = editable
        self.sortable = sortable
        self.reverse = reverse
        self.attr = attr
        self.empty = empty
        self.order_by = order_by or self.attr

    def _update_position(self):
        Formatter.__position += 1
        self.__position = Formatter.__position

    def copy(self):
        instance = copy.copy(self)
        instance._update_position()
        return instance

    @property
    def position(self):
        return self.__position

    def format(self, value, name, source):
        value = localize(value)
        return str(value)

    def extract(self, name, source, owner=None):
        if owner is not None:
            method = 'prepare_%s' % name
            func = getattr(owner, method, None)
            if func:
                return func(source)
        if self.attr:
            name = self.attr
        parts = name.split("__")
        result = source
        while parts:
            item = parts.pop(0)
            if isinstance(result, dict):
                result = result[item]
            else:
                result = getattr(result, item)
            if result is None:
                break
        return result

    def render(self, name, source, owner):
        value = self.extract(name, source, owner)
        return self.format(value, name, source)

    @classmethod
    def extract_from(cls, source, include=None, exclude=None):
        result = sorted(inspect.getmembers(source, lambda a: isinstance(a, Formatter)),
            key=lambda p: p[1].position)
        if include:
            result = [p for p in result if p[0] in list(include)]
        if exclude:
            result = [p for p in result if p[0] not in set(exclude)]
        return result


class FormattedValue(object):

    def __init__(self, name, prop, source, attrs=None, owner=None):
        self.name = name
        self.prop = prop
        self.source = source
        self.__attrs = attrs
        self.__owner = owner

    @property
    def attrs(self):
        return self.__attrs(self) if self.__attrs else {}

    @property
    def label(self):
        label = self.prop.label
        return label if label is not None else self.name.replace("_", " ").title()

    @property
    def url(self):
        if not self.prop.url:
            try:
                func = self.__owner.get_cell_url
            except AttributeError:
                return None
            else:
                return func(self)
        if callable(self.prop.url):
            return self.prop.url(self)
        return self.prop.url

    @property
    def value(self):
        try:
            return self.prop.extract(self.name, self.source, self.__owner)
        except AttributeError as ex:
            raise ValueError("Error while accessing attribute %r in %r : %s" % (self.name, self.source, ex))

    def __getattr__(self, item):
        return getattr(self.prop, item)

    def wrap_content(self, content):
        prefix = self.__owner.get_prefix(self.name)
        if prefix:
            content = "%s %s" % (prefix, content)
        suffix = self.__owner.get_suffix(self.name)
        if suffix:
            content = "%s %s" % (content, suffix)
        return Markup(content)

    def to_html(self):
        return self.wrap_content(self.to_str)

    def to_str(self):
        value = self.value
        if value is None or value == "":
            return self.prop.empty
        result = self.prop.format(value, self.name, self.source)
        return str(result)

    def __str__(self):
        return self.to_str()


class MetaFormattedObject(type):

    def __init__(cls, name, bases, attrs):
        super(MetaFormattedObject, cls).__init__(name, bases, attrs)
        cls.base_formatters = tuple((k, v) for k, v in Formatter.extract_from(cls))

    def subset_class(cls, include=None, exclude=None):
        name = "_Sub%s" % cls.__name__
        fields = Formatter.extract_from(cls, include=include, exclude=exclude)
        class_ = type(name, (cls, ), OrderedDict((k, v) for k, v in fields))
        return class_


class FormattedObject(metaclass=MetaFormattedObject):

    def __init__(self, obj, **context):
        self.context = context
        self.source = obj
        self.formatters = OrderedDict((k, v.copy()) for k,v in self.base_formatters)
        data = self.data = OrderedDict()
        for name, prop in self.formatters.items():
            data[name] = FormattedValue(name, prop, self.source, attrs=self.get_attrs, owner=self)

    def to_dict(self):
        return {c.name: str(c) for c in self}

    #This comes useful inside the templates when a single object has to be split into multiple fields
    def select(self, *names):
        stack = deque(names)
        while stack:
            item = stack.popleft()
            if isinstance(item, (list, tuple)):
                stack.extendleft(reversed(item))
            elif item in self.formatters:
                value = self[item]
                if not value.hidden:
                    yield self[item]

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __getitem__(self, item):
        try:
            return self.data[item]
        except KeyError:
            raise AttributeError(f"Error in accessing {item} of {self.source}")

    def __iter__(self):
        for name, prop in self.formatters.items():
            yield self.data[name]

    def __len__(self):
        return len(self.formatters)

    def __bool__(self):
        return len(self) > 0

    def get_attrs(self, value):
        return {}

    @classmethod
    def as_table(cls):
        props = dict(Formatter.extract_from(cls))
        for key, value in cls.__dict__.items():
            if not key.startswith("_") and callable(value):
                props[key] = value
        name = cls.__name__
        return type("%sTable"%name, (FormattedTable,), props)

    def get_prefix(self, name):
        return

    def get_suffix(self, name):
        return


class FormattedTableColumn(object):

    __inited = False

    def __init__(self, name, prop, table):
        self.name = name
        self.prop = prop
        self.table = table
        self.__inited = True

    @property
    def label(self):
        return self.prop.label or self.name.capitalize()

    def __setattr__(self, key, value):
        if self.__inited and key not in self.__dict__:
            setattr(self.prop, key, value)
        else:
            self.__dict__[key] = value

    def __getattr__(self, item):
        return getattr(self.prop, item)

    def __getitem__(self, item):
        return getattr(self.prop, item)

    def values(self):
        for obj in self:
            yield obj.value

    def __iter__(self):
        for obj in self.table.source:
            yield FormattedValue(self.name, self.prop, obj, self.table.get_cell_attrs)

    def __len__(self):
        return len(self.table.source)

    def __bool__(self):
        return len(self) > 0


class FormattedTableColumnSet(object):

    def __init__(self, table):
        self.columns = OrderedDict((n, FormattedTableColumn(n, p.copy(), table)) for n, p in table.base_formatters)

    def visible_columns(self):
        return [col for col in self.columns.values() if not col.hidden]

    def hidden_columns(self):
        return [col for col in self.columns.values() if col.hidden]

    def keys(self):
        return self.columns.keys()

    def remove(self, name):
        return self.columns.pop(name, None)

    def select(self, *names):
        clone = copy.copy(self)
        for key in self.columns.keys():
            if key not in names:
                clone.columns.pop(key)
        return clone

    def __iter__(self):
        for value in self.columns.values():
            yield value

    def __contains__(self, item):
        return item in self.columns

    def __getitem__(self, item):
        return self.columns[item]

    def __getattr__(self, item):
        return self.columns[item]

    def __len__(self):
        return len(self.columns)

    def __bool__(self):
        return len(self) > 0


class FormattedTableRow(object):

    is_footer = False

    def __init__(self, index, source, table, kind=None):
        self.index = index
        self.source = source
        self.table = table
        self.kind = kind
        self.data = OrderedDict()
        for column in self.table.columns:
            self.data[column.name] = FormattedValue(column.name, column.prop, source,
                                                    attrs=self.table.get_cell_attrs,
                                                    owner=table)

    def __getitem__(self, item):
        return self.data[item]

    @property
    def object(self):
        return self.source

    @property
    def attrs(self):
        return self.table.get_row_attrs(self)

    def __iter__(self):
        for column in self.table.columns:
            yield self.data[column.name]

    def __len__(self):
        return len(self.table.columns)

    def __bool__(self):
        return len(self) > 0


class FooterRow:

    __inited = False
    is_footer = True

    def __init__(self, table, label):
        data = self.data = {}
        for column in table.columns:
            data[column.name] = None
        self.label = label
        self.__inited = True

    def __getattr__(self, item):
        return self.data[item] if item in self.data else None

    def __setattr__(self, key, value):
        if self.__inited:
            self.data[key] = value
        else:
            self.__dict__[key] = value

    def __iter__(self):
        return iter(self.data)

    def to_row(self):
        return self.data


class FormattedTable(metaclass=MetaFormattedObject):

    def __init__(self, source, **context):
        self.context = context
        self.columns = FormattedTableColumnSet(self)
        self.source = source
        self._footer_rows = []

    def select(self, *names):
        clone = copy.copy(self)
        names = list(flatten(names))
        clone.columns = self.columns.select(names)
        return clone

    def build_links(self, request, bound_field=None):
        data = request.GET if request else {}
        sort_key = bound_field.name if bound_field else None
        sort_field = bound_field.field if bound_field else None
        for col in self.columns.visible_columns():
            if sort_key and sort_field:
                field = sort_field
                code = field.get_value_for_name(col.name)
                value = data.get(sort_key, "")
                reverse = value.startswith("-")
                if reverse:
                    value = value[1:]
                is_active = code == value
                next_value = "-%s" % code if not reverse and is_active else code
                mods = {sort_key: next_value}
            else:
                is_active = False
                reverse = False
                mods = {}
            if request:
                url = url_query_update(request.build_absolute_uri(), mods) if mods else None
            else:
                url = None
            link = Link(content=col.label, url=url, is_active=is_active, reverse=reverse, sortable=col.sortable, column=col)
            yield link

    @property
    def object_list(self):
        return self.source

    def create_footer_row(self, label):
        row = FooterRow(self, label)
        self._footer_rows.append(row)
        return row

    @property
    def footer(self):
        index = 0
        for obj in self._footer_rows:
            f = FormattedTableRow(index, obj, self)
            yield f
            index += 1

    def __iter__(self):
        index = 0
        for obj in self.source:
            yield FormattedTableRow(index, obj, self)
            index += 1

    def __len__(self):
        return len(self.source)

    def __bool__(self):
        return len(self) > 0

    def get_row_attrs(self, row):
        return {}

    def get_cell_attrs(self, cell):
        return {}

    def get_prefix(self, name):
        return

    def get_suffix(self, name):
        return





from django.contrib.staticfiles.storage import staticfiles_storage
from django.urls import reverse
from jinja2 import Environment

from .library import jinja2_filter, jinja2_function, library, Jinja2Function
from . import extensions


__all__ = ['jinja2_filter', 'jinja2_function', 'environment', 'Jinja2Function']


def inject_filters(env):
    from django.contrib.humanize.templatetags import humanize
    from django.template.defaultfilters import register

    for name, func in register.filters.items():
        if name not in env.filters:
            env.filters[name] = func

    for name, func in humanize.register.filters.items():
        if name not in env.filters:
            env.filters[name] = func


def inject_functions(env):
    pass


def inject_library(env):
    for key, func in library.functions.items():
        env.globals[key] = func

    for key, func in library.filters.items():
        env.filters[key] = func


def jinja2_reverse(view_name, **kwargs):
    extra = kwargs.pop("kwargs", {})
    kwargs.update(extra)
    return reverse(view_name, kwargs=kwargs)


def environment(**options):
    options['extensions'] = library.extensions + [extensions.PreExtension, extensions.TableExtension]
    env = Environment(**options)
    env.globals.update({
        'static': staticfiles_storage.url,
        'url': jinja2_reverse,
    })
    inject_filters(env)
    inject_functions(env)
    inject_library(env)
    return env





from django.http.request import HttpRequest
from django.core.paginator import Page as DjangoPage, Paginator as DjangoPaginator, EmptyPage, PageNotAnInteger

from djingles import utils
from djingles import html


__all__ = ["Paginator", "Page"]


class Page(DjangoPage):

    def create_link(self, request, number):
        param = self.paginator.parameter_name
        url = utils.url_query_update(request.build_absolute_uri(), {param: number})
        return html.Link(url=url, content=str(number), is_active=number==self.number)

    def build_links(self, request):
        for i in utils.page_number_range(self.number,
                                      self.paginator.page_limit,
                                      self.paginator.num_pages):
            yield self.create_link(request, i)

    def previous_link(self, request):
        number = self.previous_page_number()
        return self.create_link(request, number)

    def next_link(self, request):
        number = self.next_page_number()
        return self.create_link(request, number)


class Paginator(DjangoPaginator):

    parameter_name = "page"
    page_limit = 10
    allow_empty = False

    def __init__(self, object_list, per_page, **kwargs):
        self.parameter_name = kwargs.pop("parameter_name", self.parameter_name)
        self.allow_empty = kwargs.pop("allow_empty", self.allow_empty)
        self.page_limit = kwargs.pop("page_limit", self.page_limit)
        super(Paginator, self).__init__(object_list, per_page, **kwargs)

    def page(self, value):
        """
        Returns a Page object for the given 1-based page number.
        """
        if isinstance(value, HttpRequest):
            value = value.GET.get(self.parameter_name, 1)
        elif isinstance(value, dict):
            value = value.get(self.parameter_name, 1)
        return super(Paginator, self).page(value)

    def _get_page(self, *args, **kwargs):
        return Page(*args, **kwargs)




import inspect
from django.forms.formsets import BaseFormSet
try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_str as force_text
from django.utils.safestring import mark_safe
from djingles.jinja2 import jinja2_function
from djingles import html
from django.utils import timezone


@jinja2_function(takes_context=True, mark_safe=True)
def field_option_tags(field):
    result = []
    for choice in html.field_choices(field):
        tag = html.option(value=choice.value, selected=choice.selected)[choice.content]
        result.append(tag.render())
    return "".join(result)


@jinja2_function(takes_context=True, mark_safe=True)
def form_hidden_field_tags(context, form, csrf=False):
    request = context["request"]
    fields = []
    if csrf:
        fields.append(html.form_csrf_tag(request))
    if not isinstance(form, BaseFormSet):
        for f in form.hidden_fields():
            fields.append(str(f))
    return "".join(fields)


@jinja2_function()
def field_range(form, *args, **kwargs):
    return html.field_range(form, *args, **kwargs)


@jinja2_function()
def field_iter(form, *names):
    return html.iter_fields(form, names)


@jinja2_function(mark_safe=True)
def form_attrs(form, **kwargs):
    return html.form_attrs(form, **kwargs)


@jinja2_function(takes_context=True, mark_safe=True)
def form_start(context, form, **attrs):
    csrf = attrs.get("method", "post").lower() == "post"
    attrs = html.form_attrs(form, **attrs)
    form_tag = "<form {}>".format(attrs)
    hidden = form_hidden_field_tags(context, form, csrf=csrf)
    mgmt = force_text(getattr(form, "management_form", ""))
    return "%s%s%s" % (form_tag, hidden, mgmt)


@jinja2_function(mark_safe=True)
def form_end(form):
    return mark_safe("</form>")


@jinja2_function()
def field_choices(field):
    return html.field_choices(field)


@jinja2_function()
def field_links(request, field):
    return html.field_links(request, field)


@jinja2_function()
def field_help(field):
    return field.help_text

@jinja2_function()
def field_errors(field):
    return field.errors


@jinja2_function()
def field_label(field):
    return field.label


@jinja2_function()
def field_id(field):
    return field.id_for_label


@jinja2_function()
def field_name(field):
    return field.html_name


@jinja2_function()
def field_value(field):
    return field.value()


@jinja2_function(mark_safe=True)
def widget_tag(field, **attrs):
    return html.render_widget(field, *attrs)


@jinja2_function(mark_safe=True)
def field_tag(field, layout=None, **kwargs):
    return html.render_field(field, layout=layout, **kwargs)


@jinja2_function()
def field_class(field):
    return html.field_css_class(field)


@jinja2_function()
def form_class(form):
    return html.make_css_class(form)


@jinja2_function()
def widget_class(field):
    return html.widget_css_class(field)


@jinja2_function()
def field_is(field, class_name):
    return class_is(field.field, class_name)


@jinja2_function()
def widget_is(field, class_name):
    return class_is(field.field.widget, class_name)


@jinja2_function()
def form_visible_fields(form):
    return form.visible_fields()


@jinja2_function()
def class_is(obj, class_name):
    class_ = obj.__class__
    return class_.__name__ == class_name or any(class_name == b.__name__ for b in inspect.getmro(class_))


@jinja2_function()
def build_links(obj, request, *args, **kwargs):
    return html.build_links(obj, request, *args, **kwargs)

@jinja2_function()
def now():
    return timezone.now()


@jinja2_function()
def model_verbose_name(obj):
    return obj._meta.verbose_name.title()


@jinja2_function()
def model_verbose_name_plural(obj):
    return obj._meta.verbose_name_plural.title()



import click


@click.group()
@click.option('--debug/--no-debug', default=False)
def cli(debug):
    click.echo('Debug mode is %s' % ('on' if debug else 'off'), err=True)



if __name__ == '__main__':
    cli()



import functools
import inspect
from collections import OrderedDict
import random

from django.core.exceptions import ImproperlyConfigured, FieldDoesNotExist
from django.db.models.fields import CharField, TextField
from django.http.request import QueryDict
from django import forms
try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_str as force_text
from djingles.formatters import Formatter
from djingles.forms.fields import SortField
from djingles import utils


__all__ = ['ActionFormMixin', 'ActionModelForm', 'ActionForm', 'FilterForm',
           'FilterModelForm', 'FilterFormMixin', 'action_model_form_factory']


def _find_decorated_methods(form, attribute):
    for name, func in inspect.getmembers(form, inspect.ismethod):
        if name.startswith("_"):
            continue
        if hasattr(func, attribute):
            yield func


class ActionFormMixin(object):

    confirmation_message = None
    failure_message = None
    success_message = None

    def __init__(self, **kwargs):
        self.context = kwargs.pop("context", {})
        self.action = self.context.get("action", "execute")
        self.context["action"] = self.action
        self.action_method = getattr(self, self.action, self.execute)
        super(ActionFormMixin, self).__init__(**kwargs)
        prepare = getattr(self, 'prepare_%s' % self.action, None)
        if prepare is not None:
            prepare()

    def prepare_if_exists(self, func):
        spec = inspect.getfullargspec(func)
        names = [n for n in spec.args if n != "self"]
        try:
            args = [self.fields[f] for f in names]
        except KeyError:
            pass
        else:
            return func(*args)

    def clean_if_exists(self, func):
        spec = inspect.getfullargspec(func)
        names = [n for n in spec.args if n != "self"]
        try:
            args = [self.cleaned_data[f] for f in names]
            if any(a is None for a in args):
                return
        except KeyError:
            pass
        else:
            return func(*args)

    def __bind_fields(self):
        key = "__bound_with_form"
        for name, field in self.fields.items():
            if hasattr(field, 'bind_form') and not getattr(field, key, False):
                setattr(field, key, True)
                field.bind_form(self)

    def process_context(self, func):
        context = self.context.copy()
        if hasattr(self, "cleaned_data") and hasattr(self, "save"):
            instance = self.save(commit=False)
            context["instance"] = instance
        context["data"] = self.cleaned_data
        spec = inspect.getfullargspec(func)
        if spec.varargs:
            raise ImproperlyConfigured("Form action cannot have variable arguments")
        if spec.varkw:
            return context
        return {k: context[k] for k in spec.args[1:] if k in context}

    def process_result(self, result):
        return result

    @property
    def result(self):
        self.is_valid()
        return self.__result

    def get_success_message(self):
        if self.success_message:
            return self.success_message.format(**self.context)
        return self.success_message

    def get_failure_message(self):
        if self.failure_message:
            return self.failure_message.format(**self.context)
        return self.failure_message

    def get_confirmation_message(self):
        if self.confirmation_message:
            return self.confirmation_message.format(**self.context)
        return self.confirmation_message

    @classmethod
    def class_oid(cls):
        """
        Obfuscated class id
        :return: str
        """
        return utils.create_hash(utils.qualified_name(cls))

    @classmethod
    def call(cls, *kwargs):
        form_obj = cls(**kwargs)
        if not form_obj.is_valid():
            return False, form_obj.errors
        else:
            return True, form_obj.result

    def full_clean(self):
        result = None
        self.__bind_fields()
        super(ActionFormMixin, self).full_clean()
        try:
            _ = self.__result
        except AttributeError:
            try:
                if self.is_bound and not self._errors:
                    method = self.action_method
                    context = self.process_context(method)
                    result = method(**context)
                    result = self.process_result(result)
            except forms.ValidationError as ex:
                self.add_error(None, ex)
            finally:
                self.__result = result

    def select_fields(self, exclude=None, include=None, **extras):
        if exclude is not None:
            for f in exclude:
                self.fields.pop(f, None)
        if include is not None:
            include = set(include)
            for f in list(self.fields.keys()):
                if f not in include:
                    self.fields.pop(f)
        if extras:
            for f in extras:
                self.fields[f] = extras[f]

    def execute(self):
        return {}


class ActionForm(ActionFormMixin, forms.Form):
    pass


class ActionModelForm(ActionFormMixin, forms.ModelForm):

    confirmation_message = "Please confirm that you really wish to {action} this object ?"

    def prepare_delete(self):
        self.select_fields(include=())

    def delete(self):
        self.instance.delete()

    def update(self):
        return self.save()

    def create(self):
        return self.save()

    def clone(self):
        self.instance.id = None
        return self.save()


class FilterFormMixin:

    submit_key = None

    def __init__(self, data=None, *args, **kwargs):
        self.context = kwargs.pop("context", {})
        self.boolean_empty_label = kwargs.pop("boolean_empty_label", "Any")
        self.empty_label = kwargs.pop("empty_label", "--------------")
        unbound = False
        if not data or (self.submit_key and self.submit_key not in data):
            data = None
        super(FilterFormMixin, self).__init__(data=data, *args, **kwargs)
        if self.submit_key:
            self.fields[self.submit_key] = forms.IntegerField(
                widget=forms.HiddenInput,
                required=False
            )
        for key, f in self.fields.items():
            if f.disabled:
                continue
            if isinstance(f, forms.BooleanField) and not isinstance(f, forms.NullBooleanField):
                field = forms.NullBooleanField(required=False,
                                          label=f.label, help_text=f.help_text)
                self.fields[key] = field
                choices=((1, self.boolean_empty_label),(2, "Yes"), (3, "No"))
                field.widget.choices = choices
            if f.required:
                f.required = False
                if hasattr(f, "choices"):
                    choices = list(f.choices)
                    if choices and choices[0][0]:
                        choices.insert(0, ("", self.empty_label))
                    f.choices = tuple(choices)

    def __bind_fields(self):
        key = "__bound_with_form"
        for name, field in self.fields.items():
            if hasattr(field, 'bind_form') and not getattr(field, key, False):
                setattr(field, key, True)
                field.bind_form(self)

    @property
    def sort_field(self):
        for name, f in self.fields.items():
            if isinstance(f, SortField):
                return self[name]

    def generic_filter(self, name, queryset, value, data):
        kwargs = {}
        if isinstance(value, (tuple, list)):
            name = '%s__in' % name
        kwargs[name] = value
        return queryset.filter(**kwargs)

    def perform_filter(self, queryset):
        self.__bind_fields()
        if not self.is_bound:
            data = {k: self.get_initial_for_field(f, k) for k, f in self.fields.items()}
        else:
            if not self.is_valid():
                return self.invalid_queryset(queryset)
            data = self.cleaned_data
        allowed = set(self.fields.keys())
        if hasattr(self, "before_filters"):
            queryset = self.before_filters(queryset, data)
        for name, value in data.items():
            if name not in allowed or value in (None, '', [], ()):
                continue
            field = self.fields[name]
            if hasattr(self, "filter_%s" % name):
                result = getattr(self, "filter_%s" % name)(queryset, value, data)
            elif hasattr(field, "filter_queryset"):
                result = field.filter_queryset(queryset, value, self[name])
            else:
                result = self.generic_filter(name, queryset, value, data)
            if result is not None:
                queryset = result
        if hasattr(self, "after_filters"):
            queryset = self.after_filters(queryset, data)
        return queryset

    def invalid_queryset(self, queryset):
        return queryset.none()


class FilterForm(FilterFormMixin, forms.Form):
    pass


class FilterModelForm(FilterFormMixin, forms.ModelForm):

    def generic_filter(self, name, queryset, value, data):
        meta = queryset.model._meta
        try:
            field = meta.get_field(name)
        except FieldDoesNotExist:
            return super(FilterModelForm, self).generic_filter(name, queryset, value, data)
        else:
            op = ""
            if isinstance(value, (list, dict)):
                op = "in"
            elif isinstance(field, (CharField, TextField)):
                op = "icontains"
        query = "%s__%s" % (name, op) if op else name
        kwargs = {query: value}
        return queryset.filter(**kwargs)


def action_model_form_factory(model_class, include=None, exclude=(), **kwargs):
    name = "%sActionForm" % model_class.__name__
    meta = type("Meta", (), {"include": include, "exclude": exclude, "model": model_class})
    kwargs['Meta'] = meta
    return type(name, (ActionModelForm, ), kwargs)



from django.forms.boundfield import BoundField
from django.forms.widgets import CheckboxInput
import re
import jinja2
from collections import namedtuple
from django.middleware import csrf
try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_str as force_text
from djingles import utils
from . import common, links


__all__ = ["Choice", "form_csrf_tag", "form_attrs", "form_css_class",
           "field_choices", "field_name_range", "field_links", "iter_fields", "widget_css_class",
           "render_widget", "register_layout", "render_field", "field_css_class", "field_range",
           "wrap_csrf_token", "is_selected_choice", "make_css_class"]


Choice = namedtuple("Choice", ["name", "value", "content", "selected"])


_layouts = {}


def make_css_class(obj, suffix=""):
    name = utils.camel_to_hyphen(re.sub(r'(?i)widget|field|ginger|form|input', '', obj.__class__.__name__, 1))
    if suffix:
        name = "%s%s" % (name, suffix)
    return name


def is_selected_choice(values, choice):
    if not isinstance(values, (list, tuple)):
        values = (values, )
    text_choice = force_text(choice)
    for v in values:
        if v == choice or text_choice == force_text(v):
            return True
    return False


def field_choices(field):
    form_field = field.field
    field_value = field.value()
    name = field.html_name
    for code, label in form_field.choices:
        is_active = is_selected_choice(field_value, code)
        yield Choice(name, code, label, is_active)


def field_links(request, field):
    url = request.get_full_path()
    form_field = field.field
    field_value = field.value()
    if hasattr(form_field, 'build_links'):
        for value in form_field.build_links(request, field):
            yield value
    else:
        for code, label in form_field.choices:
            is_active = is_selected_choice(field_value, code)
            link_url = utils.url_query_update(url, {field.name: code})
            yield links.Link(link_url, label, is_active, value=code)


def form_attrs(form, **kwargs):
    attrs = kwargs
    attrs.setdefault("method", "post")
    classes = attrs.pop("class", "")
    if isinstance(classes, str):
        classes = classes.split(" ")
    classes.append(form_css_class(form))
    attrs["class"] = classes
    attrs['enctype']='multipart/form-data' if form.is_multipart() else 'application/x-www-form-urlencoded'
    return common.html_attrs(attrs)


def form_csrf_tag(request):
    csrf_token = csrf.get_token(request)
    el = common.input(type_="hidden", name="csrfmiddlewaretoken", value=csrf_token)
    return el.render()


def wrap_csrf_token(token):
    el = common.input(type_="hidden", name="csrfmiddlewaretoken", value=token)
    return el.render()


def field_range(form, start, end, step=None, hidden=True):
    field_names = field_name_range(form, start, end, step)
    return iter_fields(form, field_names, hidden=hidden)


def field_name_range(form, first, last, step=None, field_names=None):
    if field_names is None:
        field_names = list(form.fields.keys())
    keys = field_names
    if first is not None and isinstance(first, str):
        try:
            first = keys.index(first)
        except ValueError:
            raise KeyError("%r is not a field for form %r" % (first, form.__class__.__name__))
    if last is not None and isinstance(last, str):
        try:
            last = keys.index(last)-1
        except ValueError:
            raise KeyError("%r is not a field for form %r" % (last, form.__class__.__name__))
    return keys[first:last:step]


def iter_fields(form, names, hidden=True):
    for name in names:
        field = form[name]
        if hidden or not field.hidden:
            yield field


def render_field(field, layout=None, **kwargs):
    if field.is_hidden:
        return field.as_hidden()
    layout = _layouts.get(layout, default_layout)
    template = layout(field)
    ctx = {
        "field": field,
        "label": field.label,
        "label_tag": common.label(class_="form-label", for_=field.id_for_label)[field.label] if field.label else "",
        "widget": render_widget(field),
        "help": field.help_text,
        "help_tag": common.div(class_="form-help")[field.help_text],
        "errors": field.errors
    }
    content = template.format(**ctx)
    classes = ["form-field", field_css_class(field)]
    if field.errors:
        classes.append("has-error")
    return common.div(class_=classes,
                      data_field=field.name, **kwargs)[content]


def render_widget(field, **attrs):
    el = common.div(**attrs)[str(field)]
    el.attrib.update(class_=[widget_css_class(field), "form-widget"])
    return el.render()


def register_layout(name, func):
    _layouts[name] = func


def default_layout(field):
    if isinstance(field.field.widget, CheckboxInput):
        return "{widget}{label_tag}{help}{errors}"
    return "{label_tag}{widget}{help}{errors}"


def field_css_class(field):
    return make_css_class(field.field, "-field")


def widget_css_class(field):
    return make_css_class(field.field.widget, "-widget")


def form_css_class(form):
    return make_css_class(form, "-form")


def bound_field_choices(field):
    form_field = field.field
    field_value = field.value()
    name = field.html_name
    for code, label in form_field.choices:
        is_active = is_selected_choice(field_value, code)
        yield Choice(name, code, label, is_active)


def bound_field_link_builder(field, request):
    url = request.get_full_path()
    form_field = field.field
    field_value = field.value()
    if hasattr(form_field, 'build_links'):
        for value in form_field.build_links(request, field):
            yield value
    else:
        for code, label in form_field.choices:
            is_active = is_selected_choice(field_value, code)
            link_url = utils.url_query_update(url, {field.name: code})
            yield links.Link(content=label, url=link_url, is_active=is_active, value=code)


def choices_to_options(request, bound_field):
    tags = []
    for link in links.build_links(bound_field, request):
        selected = "selected" if link.is_active else ""
        html = "<option value='%s' %s> %s </option>" % (link.value, selected, link.content)
        tags.append(html)
    return jinja2.Markup("".join(tags))


links.add_link_builder(BoundField, bound_field_link_builder)






from django import forms
from django.contrib import messages
from django.forms.models import ModelForm
from django.shortcuts import redirect
from django.views.generic.base import TemplateResponseMixin, View
from djingles import utils, exceptions


__all__ = ['CommonView', 'CommonSessionDataMixin', 'CommonTemplateView', 'CommonFormView']


class CommonSessionDataMixin(object):

    session_key_prefix = ""

    def get_session_key(self):
        return "%s-%s" % (self.session_key_prefix, self.class_oid())

    def get_session_data(self):
        return self.request.session.get(self.get_session_key())

    def set_session_data(self, data):
        self.request.session[self.get_session_key()] = data

    def clear_session_data(self):
        self.request.session.pop(self.get_session_key(), None)

    @classmethod
    def class_oid(cls):
        return utils.create_hash(utils.qualified_name(cls))


class CommonView(View, CommonSessionDataMixin):

    DEBUG = messages.DEBUG
    INFO = messages.INFO
    WARNING = messages.WARNING
    ERROR = messages.ERROR
    SUCCESS = messages.SUCCESS

    def get_user(self):
        return self.request.user

    def process_request(self, request):
        self.user = self.get_user()

    def process_response(self, request, response):
        return response

    def process_exception(self, request, ex):
        if isinstance(ex, exceptions.Redirect):
            return redirect(ex.create_url(request), permanent=ex.permanent)

    def dispatch(self, request, *args, **kwargs):
        try:
            response = self.process_request(request)
            if not response:
                response = super(CommonView, self).dispatch(request, *args, **kwargs)
        except Exception as ex:
            response = self.process_exception(request, ex)
            if response is None:
                raise
        response = self.process_response(request, response)
        return response

    def get_context_data(self, **kwargs):
        kwargs.setdefault("request", self.request)
        kwargs.setdefault("user", self.request.user)
        if 'view' not in kwargs:
            kwargs['view'] = self
        return kwargs

    def add_message(self,  message, level=messages.INFO, **kwargs):
        if not message:
            return
        messages.add_message(self.request, level, message, **kwargs)

    def get_messages(self):
        return messages.get_messages(self.request)

    def get_previous_url(self, default="/"):
        return self.request.META.get("HTTP_REFERER", default)


class CommonTemplateView(CommonView, TemplateResponseMixin):

    view_icon = None

    view_label = None

    page_title = None

    page_heading = None

    page_actions = ()

    page_css_class = None

    def __init__(self, **kwargs):
        super(CommonTemplateView, self).__init__(**kwargs)
        self.extra_context = {}

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def get_page_css_class(self):
        return self.page_css_class

    def get_page_title(self):
        return self.page_title or self.get_page_heading()

    def get_page_heading(self):
        return self.page_heading

    def get_page_actions(self):
        return self.page_actions

    def get_view_icon(self):
        return self.view_icon

    def get_view_label(self):
        return self.view_label

    def render_to_response(self, ctx, **response_kwargs):
        ctx.update(self.extra_context)
        is_ajax = self.request.headers.get('x-requested-with') == 'XMLHttpRequest'
        if not is_ajax:
            ctx['view'] = self
            ctx.setdefault("messages", self.get_messages())
            ctx.setdefault('page_heading', self.get_page_heading())
            ctx.setdefault('page_title', self.get_page_title())
            ctx.setdefault('page_actions', self.get_page_actions())
            ctx.setdefault('page_css_class', self.get_page_css_class())
        response = super(CommonTemplateView, self).render_to_response(ctx, **response_kwargs)
        return response


class CommonFormView(CommonTemplateView):

    success_url = None
    form_class = None

    context_form_key = "form"

    def get_context_form_key(self, form):
        key = self.context_form_key or "form"
        return key

    def get_success_url(self, form):
        return self.success_url

    def get_form_class(self, form_key=None):
        return self.form_class

    def get_form_key(self):
        return

    def can_submit(self):
        return self.request.method == "POST"

    def post(self, request, *args, **kwargs):
        if self.can_submit():
            return self.process_submit(
                form_key=self.get_form_key(),
                data=request.POST,
                files=request.FILES
            )
        return redirect(request.get_full_path())

    def get(self, request, *args, **kwargs):
        form_key = self.get_form_key()
        if self.can_submit():
            return self.process_submit(form_key=form_key, data=request.GET)
        else:
            data, files = self.get_form_data(form_key)
            form = self.get_form(form_key=form_key, data=data, files=files)
            return self.render_form(form)

    def get_form_data(self, form_key):
        return None, None

    def set_form_data(self, form_key, data, files=None):
        raise NotImplementedError

    def form_valid(self, form):
        url = self.get_success_url(form)
        msg = form.get_success_message()
        self.add_message(level=messages.SUCCESS, message=msg)
        return redirect(url) if not callable(url) else url(form)

    def form_invalid(self, form):
        msg = form.get_failure_message()
        self.add_message(level=messages.ERROR, message=msg)
        return self.render_form(form)

    def render_form(self, form):
        form_key = self.get_form_key()
        kwargs = {self.get_context_form_key(form_key): form}
        return self.render_to_response(self.get_context_data(**kwargs))

    def get_form_initial(self, form_key):
        return None

    def get_form_prefix(self, form_key):
        return None

    def get_form_instance(self, form_key):
        return None

    def get_form_context(self, form_key):
        action = self.action
        ctx = {
            "request": self.request,
            "user": self.user,
        }
        if action:
            ctx['action'] = action
        return ctx

    def get_form_kwargs(self, form_key):
        kwargs = {
            "initial": self.get_form_initial(form_key),
            "prefix": self.get_form_prefix(form_key),
        }
        form_class = self.get_form_class(form_key)
        if form_class and issubclass(form_class, ModelForm):
            kwargs['instance'] = self.get_form_instance(form_key)
        kwargs['context'] = self.get_form_context(form_key)
        return kwargs

    def process_submit(self, form_key, data=None, files=None):
        form_obj = self.get_form(form_key, data=data, files=files)
        if form_obj.is_valid():
            return self.form_valid(form_obj)
        else:
            return self.form_invalid(form_obj)

    def get_form(self, form_key, data=None, files=None):
        form_class = self.get_form_class(form_key)
        kwargs = self.get_form_kwargs(form_key)
        kwargs['data'] = data
        kwargs['files'] = files
        form_obj = form_class(**kwargs)
        return form_obj


class CommonMultipleFormView(CommonFormView):

    form_classes = {}
    form_key_parameter = '_fkey'

    def get_context_form_key(self, form_key):
        suffix = "form"
        return suffix if not form_key else "%s_%s" % (form_key, suffix)

    def get_form_key(self):
        if not self.can_submit():
            return None
        name = self.form_key_parameter
        request = self.request
        data = request.GET if request.method == 'GET' else request.POST
        for key in self.form_classes:
            field_name = '%s-%s' % (key, name)
            if field_name in data:
                return data[field_name]

    def get_form(self, form_key, data=None, files=None):
        form_obj = super(CommonMultipleFormView, self).get_form(form_key, data, files)
        form_obj.fields[self.form_key_parameter] = forms.CharField(widget=forms.HiddenInput, initial=form_key)
        return form_obj

    def get_form_prefix(self, form_key):
        return form_key

    def get_form_class(self, form_key):
        return self.form_classes[form_key]

    def get_unbound_form_keys(self):
        return [k for k in self.form_classes.keys() if k != self.get_form_key()]

    def get(self, request, *args, **kwargs):
        key = self.get_form_key()
        if key is None:
            context = self.get_context_data(**kwargs)
            return self.render_to_response(context)
        return self.process_submit(key, data=request.GET, files=None)

    def post(self, request, *args, **kwargs):
        key = self.get_form_key()
        if key is None:
            return redirect(request.get_full_path())
        return self.process_submit(key, data=request.POST, files=request.FILES)

    def get_context_data(self, **kwargs):
        context = super(CommonMultipleFormView, self).get_context_data(**kwargs)
        for form_key in self.get_unbound_form_keys():
            data, files = self.get_form_data(form_key)
            form = self.get_form(form_key=form_key, data=data, files=files)
            context[self.get_context_form_key(form_key)] = form
        return context




from decimal import Decimal
from datetime import datetime, timedelta, date, time
from django.contrib.gis.geos.point import Point
import random
import string

from django.core.files.base import ContentFile
from . import utils


__all__ = [
    "IntegerStream",
    "FloatStream",
    "DateTimeStream",
    "DateStream",
    "TimeStream",
    "DecimalStream",
    "WordStream",
    "FirstNameStream",
    "LastNameStream",
    "FullNameStream",
    "RandomPasswordStream",
    "PasswordStream",
    "ParagraphStream",
    "SentenceStream",
    "IPAddressStream",
    "PointStream",
    "ChoiceStream",
    "ImageStream",
    "FileStream",
    "ForeignKeyStream",
    "OneToOneStream",
    "ManyToManyStream",
    "DefaultValueStream",
    "ValueStream",
    "NameStream",
    "SlugStream"
]


class IntegerStream(object):

    def __init__(self, start=0, end=255):
        self.start = start
        self.end = end

    def next(self, field):
        return random.randint(self.start, self.end)


class FloatStream(IntegerStream):

    def __init__(self, start=0, end=100.0):
        super(FloatStream, self).__init__(start, end)

    def next(self, field):
        return random.uniform(0.000001, 9999999999.99999999)


class SentenceStream(object):

    def next(self, field):
        stream = WordStream()
        result = " ".join(stream.next(field) for _ in range(random.randint(4, 9)))
        return "%s." % result.capitalize()


class ParagraphStream(object):

    def __init__(self, minimum=1, maximum=10, html=False):
        self.minimum = minimum
        self.maximum = maximum
        self.html = html

    def create_paragraph(self, field):
        stream = SentenceStream()
        return " ".join(stream.next(field) for _ in range(random.randint(5, 10)))

    def next(self, field):
        limit = random.randint(self.minimum,self.maximum)
        result = "\n".join(self.create_paragraph(field) for _ in range(limit))
        if field.max_length:
            result = result[:field.max_length]
        if self.html:
            html_result = []
            for item in result.splitlines(True):
                html_result.append('<p>'+item+'</p>')
            return ''.join(html_result)
        return result


class WordStream(object):

    def __init__(self, min_length=4, max_length=16, total=1):
        self.min_length = min_length
        self.max_length = max_length
        self.total = total

    def next(self, field):
        words = []
        for i in range(self.total):
            limit = random.randint(self.min_length, self.max_length)
            word = "".join(random.sample(string.ascii_lowercase, limit))
            words.append(word)
        return " ".join(words)


class FullNameStream(object):

    def next(self, field):
        return " ".join(NameStream().next(field) for _ in range(2))


class SlugStream(WordStream):

    def next(self, field):
        return super(SlugStream, self).next(field).replace(" ", "-")


class NameStream(object):

    def next(self, field):
        return WordStream(min_length=4, max_length=8).next(field)


class FirstNameStream(NameStream):
    pass


class LastNameStream(NameStream):
    pass


class DateTimeStream(object):

    def __init__(self, start=None, end=None, aware=True):
        if start is None:
            start = datetime(1900, 1, 1)
        if end is None:
            end = datetime.now()
        if end < start:
            raise ValueError("End %s can't be less than start %s" % (end, start))
        self.start = start
        self.end = end
        self.aware = aware

    def next(self, field):
        delta = (self.end-self.start).total_seconds()
        interval = random.randint(1, int(delta))
        result = self.start + timedelta(seconds=interval)
        if self.aware:
            from django.utils import timezone
            result = timezone.make_aware(result, timezone.get_default_timezone())
        return result


class DateStream(DateTimeStream):

    def __init__(self, start=None, end=None):
        if start is None:
            start = date(1900, 1, 1)
        if end is None:
            end = datetime.now()
        super(DateStream, self).__init__(datetime.combine(start, time()),
                                            datetime.combine(end, time()), False)

    def next(self, field):
        result = super(DateStream, self).next(field)
        return result.date()


class TimeStream(DateTimeStream):

    def __init__(self, start=None, end=None):
        if start is None:
            start = time(0, 0, 0)
        if end is None:
            end = time(23, 59, 59)
        if start > end:
            raise ValueError
        start, end = map(self.to_datetime, (start, end))
        super(TimeStream, self).__init__(start, end, False)

    @staticmethod
    def to_datetime(stamp):
        return datetime.combine(date.today(), stamp)

    def next(self, field):
        result = super(TimeStream, self).next(field)
        return result.time()


class IPAddressStream(object):

    def next(self, field):
        return ".".join(str(random.randint(1, 254)) for _ in range(4))


class ImageStream(object):
    def __init__(self, paths):
        self.folder = paths

    def next(self, field):
        filename = utils.get_random_image(self.folder)
        return ContentFile(open(filename, 'rb').read())


class FileStream(object):
    
    def __init__(self, paths, extensions=None):
        self.folders = paths
        self.extensions = extensions

    def next(self, field):
        filename = utils.get_random_file(self.folders, self.extensions)
        return ContentFile(open(filename, 'rb').read())


class DecimalStream(object):

    def __init__(self, start=None, end=None):
        if start is None:
            start = 0.0
        if end is None:
            end = 9999999999.999999
        self.start = float(start)
        self.end = float(end)

    def next(self, field):
        value = random.uniform(self.start, self.end)
        value = str(value*10**field.decimal_places)[:field.max_digits]
        return Decimal(value)/10**field.decimal_places


class PasswordStream(object):

    def __init__(self, word):
        from django.contrib.auth.hashers import make_password
        self.word = word
        self.hash_code = make_password(self.word)

    def next(self, field):
        return self.hash_code


class RandomPasswordStream(object):

    def __init__(self, max_length=32, min_length=1, specials=1, capitals=1):
        self.max_length = max_length
        self.min_length = min_length
        self.specials = specials
        self.capitals = capitals
        if max_length < min_length:
            raise ValueError("max length should be greater than min length")

    def next(self, field):
        list_special_char = ['!','@','#','$','%','^','&','*','(',')','+','_','~','<','>','|','{','}','[',']','`']
        password = ""
        size = random.randint(self.min_length, self.max_length)
        stream = WordStream()
        while len(password) < size:
            password += stream.next(field)
        password = password[:size]
        result = list(password)
        positions = set(range(len(password)))

        specials = self.specials % size
        special_positions = random.sample(range(len(password)), specials)
        positions.difference_update(special_positions)

        for i in special_positions:
            result[i] = random.choice(list_special_char)

        capitals = self.capitals % len(positions)
        for i in random.sample(positions, capitals):
            result[i] = result[i].upper()

        return "".join(result)


class ChoiceStream(object):
    def __init__(self, choices):
        self.choices = choices

    def next(self, field):
        return random.choice(self.choices)


class ValueStream(object):
    def __init__(self, value):
        self.value = value

    def next(self, field):
        return self.value


class DefaultValueStream(object):
    def next(self, field):
        return field.default


class PointStream(object):
    def __init__(self, longitude=None, latitude=None, radius=None):
        self.longitude = longitude
        self.latitude = latitude
        self.radius = radius

    def next(self, field):
        from geopy.distance import VincentyDistance
        longitude = random.randint(-179, 179) if self.longitude is None else self.longitude
        latitude = random.randint(-179, 179) if self.latitude is None else self.latitude
        radius = random.randint(100, 6000) if self.radius is None else self.radius
        d = VincentyDistance(kilometers=radius)
        bearing = random.randint(0, 359)
        p = d.destination((latitude, longitude), bearing)
        return Point(x=p.longitude, y=p.latitude)


class ForeignKeyStream(object):
    def __init__(self, queryset):
        self.queryset = queryset
        self.total = len(self.queryset)

    def next(self, field):
        i = random.randint(0, self.total-1)
        return self.queryset[i]


class OneToOneStream(ForeignKeyStream):
    pass


class ManyToManyStream(object):
    def __init__(self, queryset, limit=None):
        self.queryset = queryset
        self.total = len(self.queryset)
        self.limit = limit

    def next(self, field):
        size = self.limit if self.limit is not None else random.randint(0, self.total/2-1)
        i = random.randint(0, self.total-1)
        items = self.queryset[i: i+size]
        return items




from djingles.exceptions import LoginRequired, PermissionRequired
from django.core.exceptions import PermissionDenied


__all__ = ['PermissionRequiredMixin', 'PrivilegeRequiredMixin', 'LoginRequiredMixin',
           'StaffRequiredMixin', 'SuperUserRequiredMixin', 'OwnerRequiredMixin']


class PermissionRequiredMixin(object):

    def get_permission_url(cls, user):
        raise NotImplementedError

    def get_user(self):
        user = super(PermissionRequiredMixin, self).get_user()
        url = self.get_permission_url()
        if url is not None:
            raise PermissionRequired(url)
        return user


class LoginRequiredMixin(object):

    def get_user(self):
        user = super(LoginRequiredMixin, self).get_user()
        if not user.is_authenticated:
            raise LoginRequired
        return user


class OwnerRequiredMixin(LoginRequiredMixin):

    def get_target(self):
        obj = super(OwnerRequiredMixin, self).get_target()
        if not self.user.is_superuser and obj.owner != self.user:
            raise PermissionDenied
        return obj


class PrivilegeRequiredMixin(LoginRequiredMixin):
    
    def get_user(self):
        user = super(PrivilegeRequiredMixin, self).get_user()
        if not self.has_privileges(user):
            raise PermissionDenied
        return user

    def has_privileges(self, user):
        raise NotImplementedError


class StaffRequiredMixin(PrivilegeRequiredMixin):

    def has_privileges(self, user):
        return user.is_staff


class SuperUserRequiredMixin(PrivilegeRequiredMixin):

    def has_privileges(self, user):
        return user.is_superuser




from django.apps import AppConfig


class DjinglesConfig(AppConfig):
    name = 'djingles'

    def ready(self):
        self.scan_jinja_tags()

    def scan_jinja_tags(self):
        from djingles import utils
        from djingles.jinja2 import filters, functions
        for module in utils.iter_app_modules("jinja2globals"):
            pass




import functools
import jinja2
from django.template.loader import render_to_string
from jinja2.ext import Extension

from djingles import utils


class _Library:

    def __init__(self):
        self.functions = {}
        self.filters = {}
        self.extensions = []


library = _Library()


def jinja2_filter(name=None):
    def closure(func):
        nonlocal name
        if name is None:
            name = func.__name__
        library.filters[name] = func
        return func
    return closure


def jinja2_function(template=None, name=None, takes_context=False, mark_safe=False):
    def closure(orig_func):
        func = orig_func
        wrapper = None
        name_ = name or getattr(func,'_decorated_function', func).__name__

        if template:
            def wrapper(*args, **kwargs):
                from django.template.loader import get_template
                t = get_template(template)
                ctx = orig_func(*args, **kwargs)
                result = t.render(ctx)
                result = jinja2.Markup(result)
                return result
        elif mark_safe:
            def wrapper(*args, **kwargs):
                result = orig_func(*args, **kwargs)
                return jinja2.Markup(result)

        if wrapper:
            func = functools.update_wrapper(wrapper, func)

        if takes_context:
            func = jinja2.contextfunction(func)

        library.functions[name_] = func
        return orig_func

    return closure


class MetaJinja2Function(type):

    def __init__(self, name, bases, attrs):
        super(MetaJinja2Function, self).__init__(name, bases, attrs)
        self.jinja2_name = utils.camel_to_underscore(name)
        library.functions[self.jinja2_name] = self.as_function()


class Jinja2Function(metaclass=MetaJinja2Function):

    template_name = None

    def get_template_names(self):
        return [self.template_name]

    def get_context_data(self):
        return {}

    def render(self):
        ctx = self.get_context_data()
        ctx['me'] = self
        template_names = self.get_template_names()
        content = render_to_string(template_names, ctx)
        return jinja2.Markup(content)

    @classmethod
    def as_function(cls):
        def func(*args, **kwargs):
            instance = cls(*args, **kwargs)
            return instance.render()
        return func


class MetaJinja2Extension(type(Extension)):

    def __init__(self, name, bases, attrs):
        super().__init__(name, bases, attrs)
        library.extensions.append(self)


class Jinja2Extension(Extension, metaclass=MetaJinja2Extension):
    pass



import re
import functools
from collections import OrderedDict


__all__ = ['Link', 'add_link_builder', 'build_links']


_link_builders = {}


class Link(object):

    def __init__(self, content, icon=None, url=None, attrs=None, **kwargs):
        self.content = content
        self.icon = icon
        self.url = url
        self.patterns = map(re.compile, kwargs.pop("patterns", []))
        self.children = OrderedDict()
        self.is_active = False
        self.is_open = True
        self.attrs = attrs
        for k,v in kwargs.items():
            setattr(self, k, v)

    def set_active(self, url, prefix_match=False):
        initial = False
        if not prefix_match:
            initial = (url == self.url)
        elif prefix_match and self.patterns:
            initial = any(p.match(url) is not None for p in self.patterns)
        self.is_open = self.is_active = initial
        for n in self:
            n.set_active(url)
        if not self.is_open:
            self.is_open = any(n.is_open for n in self)
        has_active = self.is_active or any(n.is_active for n in self)
        if not has_active and not prefix_match:
            self.set_active(url, prefix_match=True)

    def __iter__(self):
        return iter(self.children.values())

    @property
    def is_leaf(self):
        return not self.children

    def add(self, content, url=None, **kwargs):
        kwargs['url'] = url
        child = Link(content, **kwargs)
        self.children[content] = child
        return child

    def __getitem__(self, item):
        return self.children[item]

    def __contains__(self, item):
        return item in self.children

    def __len__(self):
        return len(self.children)

    def child(self, path):
        fragments = path.split("/")
        folder = self
        for f in fragments:
            if f not in folder:
                folder.add(f)
            folder = folder[f]
        return folder

    def render_children(self):
        if self.children:
            return "<ul>%s</ul>"%("".join(n.render("li") for n in self),)
        else:
            return ""

    def render(self, tag="div"):
        attrs = self.html_attrs
        return "<%s><a href='%s' %s>%s</a>%s</%s>" % (tag, self.url, attrs, self.content, self.render_children(), tag)

    @property
    def html_attrs(self):
        from djingles.html import common
        attrs = common.html_attrs(self.attrs) if self.attrs else ""
        return attrs

    def build_links(self, request):
        path_info = request.path_info
        self.set_active(path_info)
        return [self]


def add_link_builder(cls, build_func):
    _link_builders[cls] = build_func


def get_link_builder(obj):
    cls = obj.__class__ if not isinstance(obj, type) else obj
    if hasattr(obj, 'build_links'):
        return obj.build_links
    else:
        try:
            func = _link_builders[cls]
        except KeyError:
            raise TypeError("No link builder associated with %r"%cls)
        else:
            return functools.partial(func, obj)


def build_links(obj, request, *args, **kwargs):
    builder = get_link_builder(obj)
    history = set()
    unique = True
    for link in builder(request, *args, **kwargs):
        if unique and link.content in history:
            continue
        history.add(link.content)
        yield link



if __name__ == '__main__':
    root = Link("Hello")
    child = root.child("Some/Things/Should/Not/Be/Seen")
    print(root.render())




from collections import OrderedDict
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from .formatters import *
from .base import FormattedObject, FormattedTable, MetaFormattedObject
from django.conf import settings


__all__ = ['FormattedModel', 'FormattedModelTable', 'object_formatter_factory', 'table_formatter_factory']


FORMATTER_MAPPING = {
    models.DateTimeField: DateTimeFormatter,
    models.DateField: DateFormatter,
    models.TimeField: TimeFormatter,
    models.ImageField: ImageFormatter,
    models.FileField: FileFormatter,
    models.FloatField: DecimalFormatter,
    models.DecimalField: DecimalFormatter
}


FORMATTER_MAPPING.update(getattr(settings, 'FIELD_FORMATTERS', {}))


def get_formatter_for_field(factory, meta, field_map, field_name, options=None):
    parts = field_name.split("__")
    field_name = parts.pop(0)
    field = field_map.get(field_name)
    if not field:
        return Formatter()
    while parts and field.model:
        model = field.related_model
        item = parts.pop(0)
        try:
            field = model._meta.get_field(item)
        except FieldDoesNotExist:
            if hasattr(field, item):
                return Formatter()
            else:
                raise

    label = field.verbose_name.title()
    label = getattr(options, 'labels', {}).get(field.name, label)

    if hasattr(factory, "formatter_for_field"):
        result = factory.formatter_for_field(field)
        if result is not None:
            result.label = label
            return result
    try:
        result = getattr(field, 'get_formatter')()
        result.label = label
        return result
    except AttributeError:
        if field.choices:
            result = ChoiceFormatter(label=label)
        else:
            result = FORMATTER_MAPPING.get(field.__class__, Formatter)(label=label)
        return result


def get_formatters_for_model(factory, model_class, fields=None, exclude=None, options=None):
    meta = model_class._meta

    extant = OrderedDict(factory.base_formatters)
    field_map = OrderedDict((f.name, f) for f in meta.get_fields() if f.concrete)

    if fields is None:
        fields = list(field_map.keys()) + [k for k in extant.keys() if k not in field_map]

    if exclude:
        exclude = set(exclude)
        fields = filter(lambda f: f not in exclude, fields)

    result = []
    for f in fields:
        if f in extant:
            fmt = extant[f]
        else:
            fmt = get_formatter_for_field(factory, meta, field_map, f, options)
        result.append((f, fmt))

    return result


class MetaFormattedModel(MetaFormattedObject):

    def __init__(cls, name, bases, attrs):
        super(MetaFormattedModel, cls).__init__(name, bases, attrs)
        try:
            if cls is FormattedModel or cls is FormattedModelTable:
                return
        except NameError:
            return
        meta = getattr(cls, 'Meta')
        model = meta.model
        base_formatters = []
        for name, formatter in get_formatters_for_model(cls, model, fields=getattr(meta, 'fields', None),
                                                        exclude=getattr(meta, 'exclude', None),
                                                        options=meta):
            # setattr(cls, name, formatter)
            base_formatters.append((name, formatter))
        cls.base_formatters = tuple(base_formatters)

    def subset_class(cls, include=None, exclude=None):
        name = "_Sub%s" % cls.__name__
        fields = cls.base_formatters
        field_names = list(include) if include is not None else [k for k, _ in fields]
        if exclude:
            field_names = [f for f in field_names if f not in set(exclude)]
        meta_class = type("Meta", (cls.Meta, ), {"fields": field_names})
        class_ = type(name, (cls, ), {"Meta": meta_class})
        return class_


class _SubsetMixin:
    pass


class FormattedModel(_SubsetMixin, FormattedObject, metaclass=MetaFormattedModel):
    pass


class FormattedModelTable(_SubsetMixin, FormattedTable, metaclass=MetaFormattedModel):
    pass


def object_formatter_factory(model_class, fields=None, exclude=None, base_class=None, **kwargs):
    name = "Formatted%sObject" % model_class.__name__
    meta = type("Meta", (), {"fields": fields, "exclude": exclude, "model": model_class})
    kwargs['Meta'] = meta
    base_class = base_class or FormattedModel
    return type(name, (base_class, ), kwargs)


def table_formatter_factory(model_class, fields=None, exclude=None, base_class=None, **kwargs):
    name = "Formatted%sTable" % model_class.__name__
    meta = type("Meta", (), {"fields": fields, "exclude": exclude, "model": model_class})
    kwargs['Meta'] = meta
    base_class = base_class or FormattedModelTable
    return type(name, (base_class, ), kwargs)



import sys
import jinja2
import os
import getpass
import shutil
import re
import click
from .main import cli


def is_venv():
    return (hasattr(sys, 'real_prefix') or
            (hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix))


def read_template(file_name, **ctx):
    file_name = os.path.realpath(os.path.join(__file__, "../templates", file_name))
    with open(file_name) as f:
        template = jinja2.Template(f.read())
        return template.render(**ctx)


@cli.command()
@click.argument("name", type=str, required=True)
@click.option("--port", type=int, required=True)
@click.option("--host", default="")
@click.option("--http", default=False)
@click.option("--threads", default=10)
@click.option("--processes", default=2)
@click.option("--env", default="")
def uwsgi_conf(name, port, http, host, threads, processes, env):
    cwd = os.getcwd()
    src = cwd if os.path.basename(cwd) == "src" else os.path.join(cwd, "src")
    base_dir = os.path.dirname(src)
    if not base_dir.endswith("/"):
        base_dir = "%s/" % base_dir
    venv = is_venv()
    ctx = locals()
    content = read_template("uwsgi.ini", **ctx)
    content = re.sub(r'\n+', '\n', content)
    click.echo(content)


@cli.command()
@click.option("--name", type=str, required=True)
@click.option("--executable", type=click.Path(exists=True, dir_okay=False), help="Path to uwsgi executable", default=lambda : shutil.which("uwsgi"))
@click.option("--conf", type=click.Path(exists=True, dir_okay=False, readable=True), required=True)
@click.option("--dir", type=click.Path(exists=True, file_okay=True), help="Working Directory", default=os.getcwd)
@click.option("--description", type=str)
@click.option("--before", type=str, default="nginx.service")
@click.option("--after", type=str, default="network.target")
@click.option("--user", type=str, default=getpass.getuser)
@click.option("--group", type=str)
def uwsgi_service(name, executable, conf, dir, description, before, after, user, group):
    ctx = locals()
    for key in ("conf", "dir", "executable"):
        ctx[key] = os.path.realpath(ctx[key])
    content = read_template("uwsgi.service", **ctx)
    content = re.sub(r'\n+', '\n', content)
    click.echo(content)


@cli.command()
@click.option("--port", default=80)
@click.option("--server-name", default="_")
@click.option("--static", default=lambda: os.path.realpath(os.path.join(os.getcwd(), "../static")))
@click.option("--media", default=lambda: os.path.realpath(os.path.join(os.getcwd(), "../media")))
@click.option("--uwsgi-port", required=True)
@click.option("--default", default=True)
def nginx_conf(port, server_name, static, media, uwsgi_port, default):
    ctx = locals()
    content = read_template("nginx.conf", **ctx)
    content = re.sub(r'\n+', '\n', content)
    click.echo(content)


from django.core.paginator import Page

from djingles.jinja2.functions import form_start, form_end
from djingles.jinja2 import jinja2_function
from djingles import html
from djingles.bootstrap4.forms import widgets


@jinja2_function()
def bs4_widget(field, **kwargs):
    return widgets.BootstrapWidget.widget_to_html(field, **kwargs)


@jinja2_function()
def bs4_field(field, layout="stack", **kwargs):
    return widgets.BootstrapWidget.field_to_html(field, layout, **kwargs)


@jinja2_function(takes_context=True, mark_safe=True)
def bs4_form_start(context, form, **attrs):
    attrs['class'] = "%s" % attrs.pop("class", "")
    return form_start(context, form, **attrs)


@jinja2_function(mark_safe=True)
def bs4_form_end(form):
    return form_end(form)


@jinja2_function(mark_safe=True)
def bs4_form_submit(form, label="Submit", icon="fa fa-save", **kwargs):
    return html.button(type="submit", class_="btn btn-primary", **kwargs)[
        html.i(class_=icon) if icon else None, " ",
        label
    ]


@jinja2_function(mark_safe=True)
def bs4_form_reset(form, label="Submit", icon="fa fa-erase", **kwargs):
    return html.button(type="reset", class_="btn btn-primary", **kwargs)[
        html.i(class_=icon) if icon else None, " ",
        label
    ]


@jinja2_function(mark_safe=True)
def bs4_form_actions(form, submit="Submit", reset=None):
    return html.div(class_="form-actions")[
        bs4_form_reset(form, label=reset) if reset else None,
        bs4_form_submit(form, label=submit) if submit else None,
    ]


@jinja2_function(mark_safe=True, template="bootstrap4/partials/pagination.html")
def bs4_pagination(page, request):
    num_pages = 0 if not isinstance(page, Page) else page.paginator.num_pages
    return {
        "num_pages": num_pages,
        "object": page,
        "request": request
    }


from urllib.parse import urljoin
from django.utils.formats import localize
from django.utils import timezone
from .base import Formatter


__all__ = ['ChoiceFormatter', 'FileFormatter', 'ImageFormatter', 'CallableFormatter',
           "DecimalFormatter", "Formatter", 'LinkFormatter', 'ExternalLinkFormatter',
           "DateFormatter", "DateTimeFormatter",
           "TimeFormatter", "PhoneFormatter", "EmailFormatter", "IntegerFormatter"]


class ChoiceFormatter(Formatter):

    def format(self, value, name, source):
        parts = name.split("__")
        tail = parts.pop()
        while parts:
            item = parts.pop(0)
            source = getattr(source, item)
        method = getattr(source, "get_%s_display" % tail, None)
        return method() if method is not None else None


class FileFormatter(Formatter):

    def format(self, value, name, source):
        if not value:
            return ""
        from django.conf import settings
        obj = value
        url = urljoin(settings.MEDIA_URL, obj.url)
        name = obj.name
        tag = "<a href='%s' target='_blank'>%s</a>" % (url, name)
        return tag


class ImageFormatter(Formatter):

    def __init__(self, **kwargs):
        self.width = kwargs.pop("width", 48)
        self.height = kwargs.pop("height", 48)
        kwargs.setdefault("variants", "detail")
        super(ImageFormatter, self).__init__(**kwargs)

    def format(self, value, name, source):
        if not value:
            return ""
        url = value.url
        tag = "<img src='%s' width='%s' height='%s' >" % (url, self.width, self.height)
        return tag


class DateTimeFormatter(Formatter):

    def format(self, value, name, source):
        if value:
            if timezone.is_naive(value):
                value = timezone.make_aware(value)
            value = timezone.localtime(value)
        return localize(value)


class DateFormatter(Formatter):

    def format(self, value, name, source):
        return localize(value)


class TimeFormatter(Formatter):

    def format(self, value, name, source):
        return localize(value)


class DecimalFormatter(Formatter):

    def __init__(self, **kwargs):
        self.decimal_places=  kwargs.pop("decimal_places", 2)
        super(DecimalFormatter, self).__init__(**kwargs)

    def format(self, value, name, source):
        template = "%%.%df" % self.decimal_places
        return template % value


class IntegerFormatter(Formatter):

    def format(self, value, name, source):
        return value


class CallableFormatter(Formatter):
    def __init__(self, func, **kwargs):
        self.func = func
        super(CallableFormatter, self).__init__(**kwargs)

    def format(self, value, name, source):
        return self.func(value)


class EmailFormatter(Formatter):

    def format(self, value, name, source):
        return "<a href='mailto:%s' > %s </a>" % (value, value)


class PhoneFormatter(Formatter):

    def format(self, value, name, source):
        return "<a href='tel:%s'> %s </a>" % (value, value)


class ExternalLinkFormatter(Formatter):

    def format(self, value, name, source):
        url = self.url
        if callable(url):
            url = url(value)
        if not url:
            return value
        return "<a href='%s' target='_blank'> %s </a>" % (url, value)


class LinkFormatter(Formatter):

    def format(self, value, name, source):
        url = self.url
        if callable(url):
            url = url(value)
        if not url:
            return value
        return "<a href='%s' > %s </a>" % (url, value)



import jinja2
import os
import getpass
import shutil
import re
import click
from .main import cli


def make_dirs(folders):
    for folder in folders:
        os.makedirs(folder, exist_ok=True)


@cli.command()
@click.argument("name")
def start_project(name):
    cwd = os.getcwd()
    src = cwd if os.path.basename(cwd) == "src" else os.path.join(cwd, "src")
    base_dir = os.path.dirname(src)
    make_dirs([os.path.join(base_dir, f) for f in ("src", "etc", "static", "media", "log", "data")])

